from datetime import datetime
from time import time

from dante.sync import Dante


def test_create_dante_on_disk(tmp_path):
    db_path = tmp_path / "test.db"
    db = Dante(db_path)
    assert db is not None
    _ = db["test"]
    assert db_path.exists()


def test_create_collection():
    db = Dante()
    coll = db["test"]
    assert coll is not None
    x = (
        db.conn.cursor()
        .execute("select name from sqlite_master where type = 'table'")
        .fetchone()
    )
    assert x == ("test",)


def test_insert_find_one():
    db = Dante()
    coll = db["test"]

    data = {"a": 1, "b": 2}
    coll.insert(data)
    result = coll.find_one(a=1, b=2)
    assert result == data


def test_insert_find_many():
    db = Dante()
    coll = db["test"]

    obj1 = {"a": 1, "b": 2, "c": 3}
    obj2 = {"a": 1, "e": 4, "f": 5}
    coll.insert(obj1)
    coll.insert(obj2)
    result = coll.find_many(a=1)
    assert len(result) == 2


def test_insert_datetime():
    db = Dante()
    coll = db["test"]

    data = {"a": 1, "b": datetime.now()}
    coll.insert(data)
    result = coll.find_one(a=1)
    assert datetime.fromisoformat(result["b"]) == data["b"]


def test_find_none():
    db = Dante()
    coll = db["test"]

    result = coll.find_one(a=1)
    assert result is None


def test_update_one():
    db = Dante()
    coll = db["test"]

    coll.insert({"a": 1, "b": 2})
    coll.update_many({"a": 1, "b": 3}, a=1)
    result = coll.find_one(a=1)
    assert result["b"] == 3


def test_update_many():
    db = Dante()
    coll = db["test"]

    coll.insert({"a": 1, "b": 2})
    coll.update_many({"a": 1, "b": 3}, a=1)
    result = coll.find_one(a=1)
    assert result["b"] == 3


def test_delete_one():
    db = Dante()
    coll = db["test"]

    coll.insert({"a": 1, "b": 2})
    coll.delete_many(a=1)
    result = coll.find_one(a=1)
    assert result is None


def test_delete_many():
    db = Dante()
    coll = db["test"]

    coll.insert({"a": 1, "b": 2})
    coll.insert({"a": 1, "b": 3})
    coll.delete_many(a=1)
    result = coll.find_many(a=1)
    assert result == []


def test_clear():
    db = Dante()
    coll = db["test"]

    coll.insert({"a": 1, "b": 2})
    coll.clear()
    result = coll.find_many()
    assert result == []


def test_insert_performance(tmp_path):
    db_path = tmp_path / "test.db"
    db = Dante(db_path, auto_commit=False)

    coll = db["test"]
    t0 = time()
    for i in range(100):
        coll.insert({"a": i, "b": i + 1})
    db.commit()
    t1 = time()
    dt = t1 - t0
    assert dt < 0.05


def test_update_performance(tmp_path):
    db_path = tmp_path / "test.db"
    db = Dante(db_path, auto_commit=False)

    coll = db["test"]
    for i in range(100):
        coll.insert({"a": i, "b": i + 1})

    db.commit()
    t0 = time()
    for i in range(100):
        coll.update_one({"a": i, "b": 2 * i + 1}, a=i)
    db.commit()
    t1 = time()
    dt = t1 - t0
    assert dt < 0.1



from __future__ import annotations

import json
import sqlite3
from typing import Iterable, TypeVar

from pydantic import BaseModel

from .base import BaseCollection, BaseDante


class Dante(BaseDante):
    def get_connection(self):
        if not self.conn:
            self.conn = sqlite3.connect(self.db_name)
        return self.conn

    def collection(self, name: str, model: BaseModel | None = None) -> "Collection":
        conn = self.get_connection()
        conn.execute(f"CREATE TABLE IF NOT EXISTS {name} (data TEXT)")
        conn.commit()
        return Collection(name, self, model)

    def commit(self):
        if self.conn:
            self.conn.commit()

    def _maybe_commit(self):
        if self.auto_commit and self.conn:
            self.commit()

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None


class Collection(BaseCollection):
    def insert(self, data: dict | BaseModel):
        cursor = self.conn.cursor()
        cursor.execute(
            f"INSERT INTO {self.name} (data) VALUES (?)", (self._to_json(data),)
        )
        self.db._maybe_commit()

    def find_many(_self, _limit=None, /, **kwargs) -> list[dict | BaseModel]:
        query, values = _self._build_query(_limit, **kwargs)

        cursor = _self.conn.cursor()
        cursor.execute(f"SELECT data FROM {_self.name}{query}", values)
        rows = cursor.fetchall()

        return [_self._from_json(row[0]) for row in rows]

    def find_one(_self, **kwargs) -> dict | BaseModel | None:
        results = _self.find_many(1, **kwargs)
        return results[0] if len(results) > 0 else None

    def update_many(_self, _data: dict | BaseModel, _limit=None, /, **kwargs):
        if not kwargs:
            raise ValueError("You must provide a filter to update")

        query, values = _self._build_query(_limit, **kwargs)

        cursor = _self.conn.cursor()
        cursor.execute(
            f"UPDATE {_self.name} SET data = ? {query}",
            (_self._to_json(_data), *values),
        )
        _self.db._maybe_commit()

    def update_one(_self, _data: dict | BaseModel, /, **kwargs):
        _self.update_many(_data, None, **kwargs)

    def delete_many(_self, _limit=None, /, **kwargs):
        if not kwargs:
            raise ValueError("You must provide a filter to delete")

        query, values = _self._build_query(_limit, **kwargs)

        cursor = _self.conn.cursor()
        cursor.execute(f"DELETE FROM {_self.name}{query}", values)
        _self.db._maybe_commit()

    def delete_one(_self, /, **kwargs):
        _self.delete_many(None, **kwargs)

    def clear(self):
        cursor = self.conn.cursor()
        cursor.execute(f"DELETE FROM {self.name}")
        self.db._maybe_commit()

    def __iter__(self) -> Iterable[dict | BaseModel]:
        return iter(self.find_many())


DanteModel = TypeVar("DanteModel", bound="DanteMixin")


class DanteMixin:
    @classmethod
    def use_db(cls: DanteModel, db: str | Dante = Dante.MEMORY):
        if hasattr(cls, "_db"):
            return

        if isinstance(db, str):
            db = Dante(db)
        cls._db = db

    @classmethod
    def close_db(cls: DanteModel):
        if hasattr(cls, "_db"):
            cls._db.close()
            del cls._db

    @classmethod
    def _get_collection(cls):
        coll = getattr(cls, "_collection", None)
        if coll:
            return coll
        coll = cls._db[cls]
        cls._collection = coll
        return coll

    @classmethod
    def find_many(cls: DanteModel, **kwargs) -> list[DanteModel]:
        return cls._get_collection().find_many(**kwargs)

    @classmethod
    def find_one(cls: DanteModel, **kwargs) -> DanteModel | None:
        return cls._get_collection().find_one(**kwargs)

    def save(self, **kwargs):
        if kwargs:
            self._get_collection().update_one(self, **kwargs)
        else:
            self._get_collection().insert(self)

    def delete(self):
        data = json.loads(self.model_dump_json())
        self._get_collection().delete_one(**data)

    @classmethod
    def delete_many(self, **kwargs):
        self._get_collection().delete_many(**kwargs)

    @classmethod
    def clear(cls: DanteModel):
        cls._get_collection().clear()



import pytest
from pydantic import BaseModel

from dante import AsyncDanteMixin


class MyModel(BaseModel):
    a: int
    b: str = "foo"


class MyDanteModel(AsyncDanteMixin, MyModel):
    pass


@pytest.mark.asyncio
async def test_get_model_collection(db):
    coll = await db[MyModel]
    assert coll.model == MyModel
    assert coll.name == "MyModel"


@pytest.mark.asyncio
async def test_insert_find_one_model(db):
    coll = await db[MyModel]

    obj = MyModel(a=1)
    await coll.insert(obj)

    result = await coll.find_one(a=1)
    assert isinstance(result, MyModel)

    assert result.a == 1
    assert result.b == "foo"


@pytest.mark.asyncio
async def test_insert_find_many_model(db):
    coll = await db[MyModel]

    obj = MyModel(a=1)
    await coll.insert(obj)

    result = await coll.find_many(a=1)
    assert len(result) == 1
    assert isinstance(result[0], MyModel)

    assert result[0].a == 1
    assert result[0].b == "foo"


@pytest.mark.asyncio
async def test_update_model(db):
    coll = await db[MyModel]

    obj = MyModel(a=1)
    await coll.insert(obj)

    obj.b = "bar"
    await coll.update_many(obj, a=1)

    result = await coll.find_one(a=1)
    assert isinstance(result, MyModel)

    assert result.a == 1
    assert result.b == "bar"


@pytest.mark.asyncio
async def test_dante_mixin(db):
    """
    Test all behavior of async DanteMixin

    This is all in one big test to avoid concurrency
    issue stemming from overriding a singleton connection
    (within DanteMixin) with a fresh connection for each test.
    """
    AsyncDanteMixin.use_db(db)

    # Test object creation and single object retrieval
    obj = MyDanteModel(a=1)
    await obj.save()

    result = await MyDanteModel.find_one(a=1)
    assert isinstance(result, MyDanteModel)

    await MyDanteModel.clear()

    # Test object update and multiple object retrieval
    obj = MyDanteModel(a=1)
    await obj.save()

    obj.b = "bar"
    await obj.save(a=1)

    results = await MyDanteModel.find_many(a=1)
    assert len(results) == 1
    assert isinstance(results[0], MyDanteModel)

    assert results[0].a == 1
    assert results[0].b == "bar"

    await MyDanteModel.clear()

    # Test object deletion (single & many)
    obj = MyDanteModel(a=1)
    await obj.save()

    await obj.delete()

    result = await MyDanteModel.find_many()
    assert result == []

    await MyDanteModel.clear()

    obj = MyDanteModel(a=1, b="foo")
    await obj.save()

    obj = MyDanteModel(a=1, b="bar")
    await obj.save()

    await MyDanteModel.delete_many(a=1)

    result = await MyDanteModel.find_many()
    assert result == []



from __future__ import annotations

import json
from typing import Iterable, TypeVar

import aiosqlite
from pydantic import BaseModel

from .base import BaseCollection, BaseDante


class Dante(BaseDante):
    async def get_connection(self) -> aiosqlite.Connection:
        if not self.conn:
            self.conn = await aiosqlite.connect(self.db_name)
        return self.conn

    async def collection(
        self,
        name: str,
        model: BaseModel | None = None,
    ) -> "Collection":
        conn = await self.get_connection()
        await conn.execute(f"CREATE TABLE IF NOT EXISTS {name} (data TEXT)")
        await conn.commit()
        return Collection(name, self, model)

    async def commit(self):
        if self.conn:
            await self.conn.commit()

    async def _maybe_commit(self):
        if self.auto_commit and self.conn:
            await self.conn.commit()

    async def close(self):
        if self.conn:
            await self.conn.close()
            self.conn = None


class Collection(BaseCollection):
    async def insert(self, data: dict | BaseModel):
        conn: aiosqlite.Connection = await self.db.get_connection()
        await conn.execute(
            f"INSERT INTO {self.name} (data) VALUES (?)", (self._to_json(data),)
        )
        await self.db._maybe_commit()

    async def find_many(_self, _limit=None, /, **kwargs) -> list[dict | BaseModel]:
        query, values = _self._build_query(_limit, **kwargs)

        conn: aiosqlite.Connection = await _self.db.get_connection()
        cursor = await conn.cursor()
        await cursor.execute(f"SELECT data FROM {_self.name}{query}", values)
        rows = await cursor.fetchall()

        return [_self._from_json(row[0]) for row in rows]

    async def find_one(_self, **kwargs) -> dict | BaseModel | None:
        results = await _self.find_many(1, **kwargs)
        return results[0] if len(results) > 0 else None

    async def update_many(_self, _data: dict | BaseModel, _limit=None, /, **kwargs):
        if not kwargs:
            raise ValueError("You must provide a filter to update")

        query, values = _self._build_query(_limit, **kwargs)

        conn: aiosqlite.Connection = await _self.db.get_connection()
        await conn.execute(
            f"UPDATE {_self.name} SET data = ? {query}",
            (_self._to_json(_data), *values),
        )
        await _self.db._maybe_commit()

    async def update_one(_self, _data: dict | BaseModel, /, **kwargs):
        await _self.update_many(_data, None, **kwargs)

    async def delete_many(_self, _limit=None, /, **kwargs):
        if not kwargs:
            raise ValueError("You must provide a filter to delete")

        query, values = _self._build_query(_limit, **kwargs)

        conn: aiosqlite.Connection = await _self.db.get_connection()
        await conn.execute(f"DELETE FROM {_self.name}{query}", values)
        await _self.db._maybe_commit()

    async def delete_one(_self, /, **kwargs):
        await _self.delete_many(None, **kwargs)

    async def clear(self):
        conn: aiosqlite.Connection = await self.db.get_connection()
        await conn.execute(f"DELETE FROM {self.name}")
        await self.db._maybe_commit()

    def __aiter__(self) -> Iterable[dict | BaseModel]:
        return iter(self.find_many())


DanteModel = TypeVar("DanteModel", bound="DanteMixin")


class DanteMixin:
    @classmethod
    def use_db(cls: DanteModel, db: str | Dante = Dante.MEMORY):
        if hasattr(cls, "_db"):
            return

        if isinstance(db, str):
            db = Dante(db)
        cls._db = db

    @classmethod
    async def close_db(cls: DanteModel):
        if hasattr(cls, "_db"):
            await cls._db.close()
            del cls._db

    @classmethod
    async def _get_collection(cls):
        coll = getattr(cls, "_collection", None)
        if coll:
            return coll
        coll = await cls._db.collection(cls.__name__, cls)
        cls._collection = coll
        return coll

    @classmethod
    async def find_many(cls: DanteModel, **kwargs) -> list[DanteModel]:
        coll = await cls._get_collection()
        return await coll.find_many(**kwargs)

    @classmethod
    async def find_one(cls: DanteModel, **kwargs) -> DanteModel | None:
        coll = await cls._get_collection()
        return await coll.find_one(**kwargs)

    async def save(self, **kwargs):
        coll = await self._get_collection()
        if kwargs:
            await coll.update_one(self, **kwargs)
        else:
            await coll.insert(self)

    async def delete(self):
        coll = await self._get_collection()
        data = json.loads(self.model_dump_json())
        await coll.delete_one(**data)

    @classmethod
    async def delete_many(self, **kwargs):
        coll = await self._get_collection()
        await coll.delete_many(**kwargs)

    @classmethod
    async def clear(cls: DanteModel):
        coll = await cls._get_collection()
        await coll.clear()



from __future__ import annotations

import json
from datetime import date, datetime

from pydantic import BaseModel


class DanteEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime) or isinstance(obj, date):
            return obj.isoformat()
        return super().default(obj)


class BaseDante:
    MEMORY = ":memory:"

    def __init__(self, db_name: str = MEMORY, auto_commit: bool = True):
        self.db_name = db_name
        self.conn = None
        self.auto_commit = auto_commit

    def __str__(self):
        return f'<{self.__class__.__name__}("{self.db_name}")>'

    def get_connection(self):
        raise NotImplementedError()

    def collection(self, name: str, model: BaseModel | None = None) -> "BaseCollection":
        raise NotImplementedError()

    def __getitem__(self, name: str | type) -> "BaseCollection":
        if isinstance(name, type) and issubclass(name, BaseModel):
            return self.collection(name.__name__, name)
        else:
            return self.collection(name)


class BaseCollection:
    def __init__(self, name: str, db: BaseDante, model: BaseModel | None = None):
        self.name = name
        self.db = db
        self.conn = db.conn
        self.model = model

    def __str__(self):
        return f'<{self.__class__.__name__}("{self.db.db_name}/{self.name}")>'

    def _to_json(self, data: dict | BaseModel) -> str:
        if self.model and isinstance(data, BaseModel):
            return data.model_dump_json()
        else:
            return json.dumps(data, cls=DanteEncoder)

    def _from_json(self, json_text: str) -> dict | BaseModel:
        data = json.loads(json_text)
        return self.model(**data) if self.model else data

    def _build_query(_self, _limit, /, **kwargs) -> tuple[str, list]:
        values = []
        if kwargs:
            query_parts = []
            for key, value in kwargs.items():
                query_parts.append("data->>? = ?")
                values.extend([key, value])
            query = " WHERE " + " AND ".join(query_parts)
        else:
            query = ""

        if _limit:
            query += f" LIMIT {_limit}"

        return query, values



from .asyncdante import Dante as AsyncDante
from .asyncdante import DanteMixin as AsyncDanteMixin
from .sync import Dante, DanteMixin

__all__ = [
    "Dante",
    "DanteMixin",
    "AsyncDante",
    "AsyncDanteMixin",
]



from pydantic import BaseModel

from dante import Dante, DanteMixin


class MyModel(BaseModel):
    a: int
    b: str = "foo"


class MyDanteModel(DanteMixin, MyModel):
    pass


def test_get_model_collection():
    db = Dante()
    coll = db[MyModel]
    assert coll.model == MyModel
    assert coll.name == "MyModel"


def test_insert_find_one_model():
    db = Dante()
    coll = db[MyModel]

    obj = MyModel(a=1)
    coll.insert(obj)

    result = coll.find_one(a=1)
    assert isinstance(result, MyModel)

    assert result.a == 1
    assert result.b == "foo"


def test_insert_find_many_model():
    db = Dante()
    coll = db[MyModel]

    obj = MyModel(a=1)
    coll.insert(obj)

    result = coll.find_many(a=1)
    assert len(result) == 1
    assert isinstance(result[0], MyModel)

    assert result[0].a == 1
    assert result[0].b == "foo"


def test_update_model():
    db = Dante()
    coll = db[MyModel]

    obj = MyModel(a=1)
    coll.insert(obj)

    obj.b = "bar"
    coll.update_many(obj, a=1)

    result = coll.find_one(a=1)
    assert isinstance(result, MyModel)

    assert result.a == 1
    assert result.b == "bar"


def test_dante_model_insert():
    DanteMixin.use_db()
    MyDanteModel.clear()

    obj = MyDanteModel(a=1)
    obj.save()

    result = MyDanteModel.find_one(a=1)
    assert isinstance(result, MyDanteModel)


def test_dante_model_update():
    DanteMixin.use_db()
    MyDanteModel.clear()

    obj = MyDanteModel(a=1)
    obj.save()

    obj.b = "bar"
    obj.save(a=1)

    results = MyDanteModel.find_many(a=1)
    assert len(results) == 1
    assert isinstance(results[0], MyDanteModel)

    assert results[0].a == 1
    assert results[0].b == "bar"


def test_dante_model_delete():
    DanteMixin.use_db()
    MyDanteModel.clear()

    obj = MyDanteModel(a=1)
    obj.save()

    obj.delete()

    result = MyDanteModel.find_many()
    assert result == []


def test_dante_model_delete_many():
    DanteMixin.use_db()
    MyDanteModel.clear()

    obj = MyDanteModel(a=1, b="foo")
    obj.save()

    obj = MyDanteModel(a=1, b="bar")
    obj.save()

    MyDanteModel.delete_many(a=1)

    result = MyDanteModel.find_many()
    assert result == []
